/*
 * trickle.h
 *
 * Copyright (c) 2003 Marius Aamodt Eriksen <marius@monkey.org>
 * All rights reserved.
 *
 * $Id: trickle.h,v 1.3 2003/04/06 00:29:37 marius Exp $
 */

#ifndef TRICKLE_H
#define TRICKLE_H

#define TRICKLE_SEND 0
#define TRICKLE_RECV 1

#define TRICKLE_NONBLOCK 0x1

#define TRICKLE_WOULDBLOCK 1

//each division is 5 minutes
#define DIVS_PER_HOUR 12

#endif /* !TRICKLE_H */
